#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <TROOT.h>
#include <TMath.h>
#include <TF1.h>
#include <TH1F.h>
#include <TGraphErrors.h>
#include <TFile.h>
#include <TString.h>
#include <TStopwatch.h>

Double_t ftMeanN(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t n0 = par[0];
    Double_t a1 = par[1];
    Double_t a2 = par[2];
    Double_t a3 = par[3];
    // Variables
    Double_t xx = x[0];
    // Function
    Double_t func = n0 * TMath::Exp(-a1 * xx - a2 * xx * xx - a3 * xx * xx * xx);

    return func;
}

Double_t ftSigmN(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t sig0 = par[0];
    Double_t n0 = par[1];
    Double_t a1 = par[2];
    Double_t a2 = par[3];
    Double_t a3 = par[4];
    // Variables
    Double_t xx = x[0];
    // Function
    Double_t mean_n = ftMeanN(&x[0], &par[1]);
    Double_t func = sig0 * TMath::Sqrt(mean_n / n0);

    return func;
}

Double_t ftEtaN(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t sig0 = par[0];
    Double_t n0 = par[1];
    Double_t a1 = par[2];
    Double_t a2 = par[3];
    Double_t a3 = par[4];
    // Variables
    Double_t xx = x[0];
    // Function
    Double_t mean_n = ftMeanN(&x[0], &par[1]);
    Double_t sigma = ftSigmN(&x[0], &par[0]);
    Double_t func = 2 * TMath::Power(1 + TMath::Erf(mean_n / (TMath::Sqrt(2) * sigma)), -1);

    return func;
}

Double_t ftProbIntegrand(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t n = par[0];
    Double_t sig0 = par[1];
    Double_t n0 = par[2];
    Double_t a1 = par[3];
    Double_t a2 = par[4];
    Double_t a3 = par[5];
    // Variables
    Double_t xx = x[0];
    // Function
    Double_t mean_n = ftMeanN(&x[0], &par[2]);
    Double_t sigma = ftSigmN(&x[0], &par[1]);
    Double_t func = 1 / (sigma * TMath::Sqrt(2 * TMath::Pi())) *
                    TMath::Exp(-TMath::Power((n - mean_n), 2) / (2 * sigma * sigma));

    return func;
}

Double_t ftProbIntegrandEta(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t n = par[0];
    Double_t sig0 = par[1];
    Double_t n0 = par[2];
    Double_t a1 = par[3];
    Double_t a2 = par[4];
    Double_t a3 = par[5];
    // Variables
    Double_t xx = x[0];
    // Function
    Double_t mean_n = ftMeanN(&x[0], &par[2]);
    Double_t sigma = ftSigmN(&x[0], &par[1]);
    Double_t eta = ftEtaN(&x[0], &par[1]);
    Double_t func = eta / (sigma * TMath::Sqrt(2 * TMath::Pi())) *
                    TMath::Exp(-TMath::Power((n - mean_n), 2) / (2 * sigma * sigma));

    return func;
}

Double_t ftProbN(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t sig0 = par[0];
    Double_t n0 = par[1];
    Double_t a1 = par[2];
    Double_t a2 = par[3];
    Double_t a3 = par[4];
    // Variables
    Double_t n = x[0];
    // Function
    TF1 *f = new TF1("f", ftProbIntegrand, 0, 1, 6);
    f->SetParameters(n, sig0, n0, a1, a2, a3);
    Double_t func = f->Integral(0, 1);

    return func;
}

Double_t ftProbNEta(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t sig0 = par[0];
    Double_t n0 = par[1];
    Double_t a1 = par[2];
    Double_t a2 = par[3];
    Double_t a3 = par[4];
    // Variables
    Double_t n = x[0];
    // Function
    TF1 *f = new TF1("f", ftProbIntegrandEta, 0, 1, 6);
    f->SetParameters(n, sig0, n0, a1, a2, a3);
    Double_t func = f->Integral(0, 1);

    return func;
}

Double_t ftCentBIntegrand(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t n = par[0];
    Double_t sig0 = par[1];
    Double_t n0 = par[2];
    Double_t a1 = par[3];
    Double_t a2 = par[4];
    Double_t a3 = par[5];
    // Variables
    Double_t xx = x[0];
    // Function
    Double_t mean_n = ftMeanN(&x[0], &par[2]);
    Double_t sigma = ftSigmN(&x[0], &par[1]);
    Double_t func = 0.5 * TMath::Erfc((n - mean_n) / (TMath::Sqrt(2) * sigma));

    return func;
}

Double_t ftCentB(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t sig0 = par[0];
    Double_t n0 = par[1];
    Double_t a1 = par[2];
    Double_t a2 = par[3];
    Double_t a3 = par[4];
    // Variables
    Double_t n = x[0];
    // Function
    TF1 *f = new TF1("f", ftCentBIntegrand, 0, 1, 6);
    f->SetParameters(n, sig0, n0, a1, a2, a3);
    Double_t func = f->Integral(0, 1);

    return func;
}

Double_t ftProbCb(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t n = par[0];
    Double_t sig0 = par[1];
    Double_t n0 = par[2];
    Double_t a1 = par[3];
    Double_t a2 = par[4];
    Double_t a3 = par[5];
    // Variables
    Double_t xx = x[0];
    // Function
    TF1 *f = new TF1("f", ftProbIntegrand, 0, 1, 6);
    f->SetParameters(n, sig0, n0, a1, a2, a3);
    Double_t Denom = f->Integral(0, 1);
    Double_t Nom = f->Eval(xx);
    Double_t func = Nom / Denom;

    return func;
}

Double_t ftProbB(Double_t *x, Double_t *par)
{
    // Parameters
    Double_t sigInel = par[0];
    Double_t n = par[1];
    Double_t sig0 = par[2];
    Double_t n0 = par[3];
    Double_t a1 = par[4];
    Double_t a2 = par[5];
    Double_t a3 = par[6];
    // Variables
    Double_t b = x[0];
    Double_t cb = TMath::Pi() * TMath::Power(b, 2) / sigInel;
    // Function
    TF1 *f = new TF1("f", ftProbCb, 0, 1, 6);
    f->SetParameters(n, sig0, n0, a1, a2, a3);
    Double_t func = 2 * TMath::Pi() * b / sigInel * f->Eval(cb);

    return func;
}

int main(int argc, char **argv)
{
    if (argc != 5)
    {
        std::cerr << "Error: wrong number of arguments:" << std::endl;
        std::cerr << "./FitCent -dat/-root inputfile -o output" << std::endl;
        return 10;
    }

    TStopwatch timer;
    timer.Start();

    // Number of iterations during fitting procedure
    int NumOfIterations = 5;

    // Reference statistics for hBimp
    int RefStat = 1e6;

    const TString centName[10] = {"0-10%", "10-20%", "20-30%", "30-40%",
                                  "40-50%", "50-60%", "60-70%", "70-80%",
                                  "80-90%", "90-100%"};

    // Parsing arguments
    TString inputfile, outputfile;
    Bool_t isRoot = false;
    for (int i = 1; i < argc; i++)
    {
        if (std::string(argv[i]) != "-dat" &&
            std::string(argv[i]) != "-root" &&
            std::string(argv[i]) != "-o")
        {
            std::cerr << "Error: Unknown option: " << argv[i] << std::endl;
            return 11;
        }
        else
        {
            if ((std::string(argv[i]) == "-dat" || std::string(argv[i]) == "-dat") && i == argc - 1)
            {
                std::cerr << "Error: input file was not specified!" << std::endl;
                return 12;
            }
            if (std::string(argv[i]) == "-root" && i != argc - 1)
            {
                inputfile = argv[++i];
                isRoot = true;
            }
            if (std::string(argv[i]) == "-dat" && i != argc - 1)
            {
                inputfile = argv[++i];
                isRoot = false;
            }
            if (std::string(argv[i]) == "-o" && i == argc - 1)
            {
                std::cerr << "Error: output file was not specified!" << std::endl;
                return 13;
            }
            if (std::string(argv[i]) == "-o" && i != argc - 1)
            {
                outputfile = argv[++i];
            }
        }
    }

    std::cout << "Input file:  " << inputfile << std::endl;
    std::cout << "Output file: " << outputfile << std::endl;
    std::vector<Double_t> vMult, vProb, vError, vErrorX;
    Double_t mult, prob, error;
    int counter = 0;
    if (!isRoot)
    {
        std::ifstream infile(inputfile.Data());
        if (!infile.is_open())
        {
            std::cerr << "Error: can not open the file!" << std::endl;
            return 20;
        }
        std::string str;
        const Double_t bin_width = 0.44;
        while (std::getline(infile, str))
        {
            std::istringstream iss(str);
            if (!(iss >> mult >> prob))
            {
                std::cerr << "Error: can not read line " << counter << ": " << str << std::endl;
                return 30;
            }

            error = prob / (1e7 * 0.44);

            vMult.push_back(mult);
            vProb.push_back(prob);
            vError.push_back(error);
            vErrorX.push_back(0.);

            counter++;
        }
        std::cout << counter << " lines were processed." << std::endl;
    }
    if (isRoot)
    {
        TFile *fi = new TFile(inputfile.Data(), "read");
        TH1F *hist = (TH1F *)fi->Get("hRefMultSTAR");
        if (!hist)
        {
            std::cerr << "Could not read histogram from root-file!" << std::endl;
            return 31;
        }

        for (int i = 0; i < hist->GetNbinsX(); i++)
        {
            if (hist->GetBinContent(i + 1) == 0)
                continue;
            vMult.push_back(hist->GetBinCenter(i + 1));
            vProb.push_back(hist->GetBinContent(i + 1) / hist->Integral());
            vError.push_back(hist->GetBinError(i + 1) / hist->Integral());
            vErrorX.push_back(0.);
            counter++;
        }
        std::cout << counter << " bins were processed." << std::endl;
        delete hist;
        fi->Close();
    }

    // Getting input data into TGraphErrors
    TGraphErrors *grMult = new TGraphErrors(vMult.size(), &vMult[0], &vProb[0], &vErrorX[0], &vError[0]);
    grMult->SetName("grMult");
    grMult->GetYaxis()->SetTitle("1/N dN/dN_{ch}");
    grMult->GetXaxis()->SetTitle("N_{ch}");

    // Fitting procedure
    TF1 *fitFunc = new TF1("fitFunc", ftProbN, vMult.at(25), vMult.at(vMult.size() - 1), 5);
    fitFunc->SetParameters(6., 100., 5., -2., 5.);

    for (int iter = 0; iter < NumOfIterations; iter++)
    {
        std::cout << "Fit iteration: " << iter << std::endl;
        grMult->Fit(fitFunc, "RM");
    }

    // Fitting function w/ ftEtaN
    TF1 *fitEtaFunc = new TF1("fitEtaFunc", ftProbNEta, vMult.at(0), vMult.at(vMult.size() - 1), 5);
    for (int ipar = 0; ipar < 5; ipar++)
    {
        fitEtaFunc->SetParameter(ipar, fitFunc->GetParameter(ipar));
    }

    // Build Centrality vs N_{ch} function using fit parameters
    TF1 *centFunc = new TF1("centFunc", ftCentB, 0, vMult.at(vMult.size() - 1), 5);
    for (int ipar = 0; ipar < 5; ipar++)
    {
        centFunc->SetParameter(ipar, fitFunc->GetParameter(ipar));
    }

    // Calculate probability functions for 10% step centrality
    TF1 *probFunc[10];
    TF1 *probFuncB[10];
    TF1 *funcTmp;
    TH1F *hBimp[10];
    TH1F *hBimpTotal = new TH1F(Form("hBimpTotal"), Form("%s","0-100%"), 200, 0, 20);
    Double_t n_test, n1, n2;
    const Double_t sigma_Inelastic = 685.; // [fm^2] for Au+Au
    // const Double_t sigma_Inelastic = 777.; // [fm^2] for Pb+Pb
    for (int icent = 0; icent < 10; icent++)
    {
        // std::cout << "CHECKIN' " << icent << std::endl;
        probFunc[icent] = new TF1(Form("probFunc_cent%i", icent), ftProbCb, 0, 1, 6);
        probFuncB[icent] = new TF1(Form("probFuncB_cent%i", icent), ftProbB, 0, 20, 7);
        hBimp[icent] = new TH1F(Form("hBimp_cent%i", icent), Form("%s", centName[icent].Data()), 200, 0, 20);
        for (int jcent = 0; jcent < 10; jcent++)
        {
            // std::cout << "\tCentrality = " << icent*10+jcent << std::endl;
            if ((icent + jcent) == 0 || (icent * 10 + jcent) >= 98)
                continue;
            n_test = centFunc->GetX(icent / 10. + jcent / 100.);
            funcTmp = new TF1(Form("funcTmp"), ftProbB, 0, 20, 7);
            funcTmp->SetParameter(0, sigma_Inelastic);
            funcTmp->SetParameter(1, n_test);
            for (int ipar = 0; ipar < 5; ipar++)
            {
                funcTmp->SetParameter(ipar + 2, fitFunc->GetParameter(ipar));
            }
            hBimp[icent]->FillRandom("funcTmp", RefStat);
            hBimpTotal->FillRandom("funcTmp", RefStat);
            delete funcTmp;
        }
        n_test = centFunc->GetX(icent / 10. + 0.05);
        probFunc[icent]->SetParameter(0, n_test);
        probFuncB[icent]->SetParameter(0, sigma_Inelastic);
        probFuncB[icent]->SetParameter(1, n_test);
        for (int ipar = 0; ipar < 5; ipar++)
        {
            probFunc[icent]->SetParameter(ipar + 1, fitFunc->GetParameter(ipar));
            probFuncB[icent]->SetParameter(ipar + 2, fitFunc->GetParameter(ipar));
        }
    }

    std::cout << std::endl;
    std::cout << "Writing output." << std::endl;

    TFile *fo = new TFile(outputfile.Data(), "recreate");
    fo->cd();

    grMult->Write();
    fitFunc->Write();
    fitEtaFunc->Write();
    centFunc->Write();
    for (int icent = 0; icent < 10; icent++)
    {
        probFunc[icent]->Write();
    }
    for (int icent = 0; icent < 10; icent++)
    {
        probFuncB[icent]->Write();
    }
    for (int icent = 0; icent < 10; icent++)
    {
        hBimp[icent]->Write();
    }
    hBimpTotal->Write();

    fo->Close();

    timer.Stop();
    timer.Print();
    return 0;
}